/**
 * Copyright 2010-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.protocol.ssh.impl.jsch;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.HashMap;

import org.ow2.util.file.FileUtils;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.protocol.api.ITransferProvider;
import org.ow2.util.protocol.api.ProtocolException;
import org.ow2.util.protocol.ssh.api.SshProtocolConfig;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.JSchException;

/**
 *
 * @author Loris Bouzonnet
 */
public class ScpTransferProvider implements ITransferProvider {

	private static Log logger = LogFactory.getLog(ScpTransferProvider.class);

	private final SshProtocolSession protocolSession;

	public ScpTransferProvider(final SshProtocolSession protocolSession) {
		this.protocolSession = protocolSession;
	}

	public void download(final String source, final String destination, boolean recursive)
	throws ProtocolException {

		final String rfile;
		final SshProtocolSession protocolSession;

		if (this.protocolSession == null) {
			final String url = source.substring(0, source.indexOf("/"));
			rfile =
				source.substring(source.indexOf("/"));
			protocolSession = new SshProtocol().createSession(
					SshProtocolConfigFactory.createSshProtocolConfig(url, null));
		} else {
			rfile = source;
			protocolSession = this.protocolSession;
		}


		String prefix = null;
		if (new File(destination).isDirectory()) {
			prefix = destination + File.separator;
		}

		// exec 'scp -f rfile' remotely
		final String command = "scp -f " + rfile;

		FileOutputStream fos = null;
		Channel channel = null;
		try {
			channel =
				protocolSession.createJschChannel("exec", command);

			// get I/O streams for remote scp
			final OutputStream out = channel.getOutputStream();
			final InputStream in = channel.getInputStream();

			channel.connect();

			final byte[] buf = new byte[1024];

			// send '\0'
			buf[0] = 0;
			out.write(buf, 0, 1);
			out.flush();

			while (true) {
				final int c = checkAck(in);
				if (c != 'C') {
					break;
				}

				// read '0644 '
				in.read(buf, 0, 5);

				long filesize = 0L;
				while (true) {
					if (in.read(buf, 0, 1) < 0) {
						// error
						break;
					}
					if (buf[0] == ' ') {
						break;
					}
					filesize = filesize * 10L + (buf[0] - '0');
				}

				String file = null;
				for (int i = 0;; i++) {
					in.read(buf, i, 1);
					if (buf[i] == (byte) 0x0a) {
						file = new String(buf, 0, i);
						break;
					}
				}

				// send '\0'
				buf[0] = 0;
				out.write(buf, 0, 1);
				out.flush();

				// read a content of lfile
				fos =
					new FileOutputStream(prefix == null ? destination : prefix
							+ file);
				int foo;
				while (true) {
					if (buf.length < filesize) {
						foo = buf.length;
					} else {
						foo = (int) filesize;
					}
					foo = in.read(buf, 0, foo);
					if (foo < 0) {
						// error
						break;
					}
					fos.write(buf, 0, foo);
					filesize -= foo;
					if (filesize == 0L) {
						break;
					}
				}
				fos.close();
				fos = null;

				if (checkAck(in) != 0) {
					return;
				}

				// send '\0'
				buf[0] = 0;
				out.write(buf, 0, 1);
				out.flush();
			}

		} catch (final IOException e) {
			throw new ProtocolException(e);
		} catch (final JSchException e) {
			throw new ProtocolException(e);
		} finally {
			if (channel != null) {
				channel.disconnect();
			}
			if (this.protocolSession == null) {
				FileUtils.close(protocolSession);
			}
		}
	}

	public void upload(final String source, final String destination, boolean recursive)
	throws ProtocolException {


		final String rfile;
		final SshProtocolSession protocolSession;

		if (this.protocolSession == null) {
			final String url = destination.substring(0, destination.indexOf("/"));
			rfile =
				destination.substring(destination.indexOf("/"));
			protocolSession = new SshProtocol().createSession(
					SshProtocolConfigFactory.createSshProtocolConfig(url, null));
		} else {
			rfile = destination;
			protocolSession = this.protocolSession;
		}

		// exec 'scp -t rfile' remotely
		String command = "scp -p -t " + rfile;

		FileInputStream fis = null;
		Channel channel = null;
		try {
			channel =
				protocolSession.createJschChannel("exec", command);

			// get I/O streams for remote scp
			final OutputStream out = channel.getOutputStream();
			final InputStream in = channel.getInputStream();

			channel.connect();

			if (checkAck(in) != 0) {
				return;
			}

			// send "C0644 filesize filename", where filename should not include
			// '/'
			final long filesize = (new File(source)).length();
			command = "C0644 " + filesize + " ";
			if (source.lastIndexOf(File.separatorChar) > 0) {
				command += source.substring(source.lastIndexOf(File.separatorChar) + 1);
			} else {
				command += source;
			}
			command += "\n";
			out.write(command.getBytes());
			out.flush();
			if (checkAck(in) != 0) {
				return;
			}

			// send a content of lfile
			fis = new FileInputStream(source);
			final byte[] buf = new byte[1024];
			while (true) {
				final int len = fis.read(buf, 0, buf.length);
				if (len <= 0) {
					break;
				}
				out.write(buf, 0, len); // out.flush();
			}
			fis.close();
			fis = null;
			// send '\0'
			buf[0] = 0;
			out.write(buf, 0, 1);
			out.flush();
			if (checkAck(in) != 0) {
				return;
			}
			out.close();

		} catch (final IOException e) {
			throw new ProtocolException(e);
		} catch (final JSchException e) {
			throw new ProtocolException(e);
		} finally {
			if (channel != null) {
				channel.disconnect();
			}
			if (this.protocolSession == null) {
				FileUtils.close(protocolSession);
			}
		}
	}

	private int checkAck(final InputStream in) throws IOException {
		final int b = in.read();
		// b may be 0 for success,
		// 1 for error,
		// 2 for fatal error,
		// -1
		if (b == 0) {
			return b;
		}
		if (b == -1) {
			return b;
		}

		if (b == 1 || b == 2) {
			final StringBuffer sb = new StringBuffer();
			int c;
			do {
				c = in.read();
				sb.append((char) c);
			} while (c != '\n');
			if (b == 1) { // error
				logger.error(sb.toString());
			}
			if (b == 2) { // fatal error
				logger.fatal(sb.toString());
			}
		}
		return b;
	}

	public static void main(final String[] arg) throws ProtocolException {

		if (arg.length!=2){
			throw new ProtocolException();
		}

		final String url;
		String downloadFile = null;
		String uploadFile = null;

		if (arg[0].contains(":")) {
			url = arg[0].substring(0, arg[0].indexOf("/"));
			downloadFile = arg[0].substring(arg[0].indexOf("/"));
		} else if (arg[1].contains(":")) {
			url = arg[1].substring(0, arg[1].indexOf("/"));
			uploadFile = arg[1].substring(arg[1].indexOf("/"));
		} else {
			throw new ProtocolException();
		}

		final HashMap<String, String> context = new HashMap<String, String>();
		context.put(SshProtocolConfigFactory.PRIVATE_KEY_URL,
				System.getProperty("user.home") + File.separator + ".ssh" + File.separator + "id_rsa");
		context.put(SshProtocolConfigFactory.KNOWN_HOSTS,
				System.getProperty("user.home") + File.separator + ".ssh" + File.separator + "known_hosts");

		final SshProtocolConfig protocolConfig =
			SshProtocolConfigFactory.createSshProtocolConfig(url, context);

		final SshProtocolSession session = new SshProtocol().createSession(protocolConfig);
		final ScpTransferProvider scpTransfertProvider = new ScpTransferProvider(session);

		if (downloadFile != null) {
			scpTransfertProvider.download(downloadFile, arg[1], false);
		} else {
			scpTransfertProvider.upload(arg[0], uploadFile, false);
		}

		FileUtils.close(session);
	}

}
