/**
 * Copyright 2010-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.protocol.ssh.impl.jsch;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.HashMap;

import org.ow2.util.file.FileUtils;
import org.ow2.util.protocol.api.IShell;
import org.ow2.util.protocol.api.ProtocolConfig;
import org.ow2.util.protocol.api.ProtocolException;
import org.ow2.util.protocol.ssh.api.SshProtocolConfig;

import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.SftpATTRS;
import com.jcraft.jsch.SftpException;

/**
 * Based on the Sftp example provided by Jsch.
 * @author Loris Bouzonnet
 */
public class SftpShell implements IShell {

    private final SshProtocolSession protocolSession;

    private ChannelSftp channel;

    private final PrintStream out;

    private final PrintStream err;

    public SftpShell(final SshProtocolSession protocolSession)
    throws ProtocolException {
        this.protocolSession = protocolSession;
        final ProtocolConfig config = protocolSession.getConfig();
        if (config.getOutputStream() == null) {
            out = System.out;
        } else if (config.getOutputStream() instanceof PrintStream) {
            out = (PrintStream) config.getOutputStream();
        } else {
            out = new PrintStream(config.getOutputStream());
        }
        if (config.getErrorStream() == null) {
            err = System.err;
        } else if (config.getErrorStream() instanceof PrintStream) {
            err = (PrintStream) config.getErrorStream();
        } else {
            err = new PrintStream(config.getErrorStream());
        }
        init();
    }

    private void init() throws ProtocolException {

        channel = (ChannelSftp) protocolSession.createJschChannel("sftp", null);
        try {
            channel.connect();
        } catch (final JSchException e) {
            throw new ProtocolException(e);
        }
    }

    public void close() throws IOException {
        channel.disconnect();
    }

    public void execute(final String command) throws ProtocolException {

        final Session session = protocolSession.getSession();
        final String[] cmds = command.split(" ");
        final String cmd = cmds[0];

        if (cmd.equals("quit")) {
            channel.quit();
        } else if (cmd.equals("exit")) {
            channel.exit();
        } else if (cmd.equals("rekey")) {
            try {
                session.rekey();
            } catch (final Exception e) {
                err.println(e);
            }
        } else if (cmd.equals("cd") || cmd.equals("lcd")) {
            if (cmds.length != 2) {
                err.println("Invalid arguments");
                return;
            }
            final String path = cmds[1];
            try {
                if (cmd.equals("cd")) {
                    channel.cd(path);
                } else {
                    channel.lcd(path);
                }
            } catch (final SftpException e) {
                err.println(e);
            }
        } else if (cmd.equals("rm") || cmd.equals("rmdir")
                || cmd.equals("mkdir")) {
            if (cmds.length != 2) {
                err.println("Invalid arguments");
                return;
            }
            final String path = cmds[1];
            try {
                if (cmd.equals("rm")) {
                    channel.rm(path);
                } else if (cmd.equals("rmdir")) {
                    channel.rmdir(path);
                } else {
                    channel.mkdir(path);
                }
            } catch (final SftpException e) {
                err.println(e);
            }
        } else if (cmd.equals("chgrp") || cmd.equals("chown")
                || cmd.equals("chmod")) {
            if (cmds.length != 3) {
                err.println("Invalid arguments");
                return;
            }
            final String path = cmds[2];
            int foo = 0;
            if (cmd.equals("chmod")) {
                final byte[] bar = cmds[1].getBytes();
                int k;
                for (int j = 0; j < bar.length; j++) {
                    k = bar[j];
                    if (k < '0' || k > '7') {
                        foo = -1;
                        break;
                    }
                    foo <<= 3;
                    foo |= (k - '0');
                }
                if (foo == -1) {
                    throw new ProtocolException();
                }
            } else {
                try {
                    foo = Integer.parseInt(cmds[1]);
                } catch (final Exception e) {
                    throw new ProtocolException(e);
                }
            }
            try {
                if (cmd.equals("chgrp")) {
                    channel.chgrp(foo, path);
                } else if (cmd.equals("chown")) {
                    channel.chown(foo, path);
                } else if (cmd.equals("chmod")) {
                    channel.chmod(foo, path);
                }
            } catch (final SftpException e) {
                err.println(e);
            }
        } else if(cmd.equals("pwd") || cmd.equals("lpwd")){
            String str=(cmd.equals("pwd")?"Remote":"Local");
            str+=" working directory: ";
            if(cmd.equals("pwd")) {
                try {
                    str+=channel.pwd();
                } catch (final SftpException e) {
                    err.println(e);
                }
            } else {
                str+=channel.lpwd();
            }
            out.println(str);
        } else if(cmd.equals("ls") || cmd.equals("dir")){
            String path=".";
            if(cmds.length==2) {
                path=cmds[1];
            }
            try{
                final java.util.Vector<?> vv=channel.ls(path);
                if(vv!=null){
                    for(int ii=0; ii<vv.size(); ii++){

                        final Object obj=vv.elementAt(ii);
                        if(obj instanceof com.jcraft.jsch.ChannelSftp.LsEntry){
                            out.println(((com.jcraft.jsch.ChannelSftp.LsEntry)obj).getLongname());
                        }

                    }
                }
            } catch(final SftpException e){
                err.println(e);
            }
        } else if(cmd.equals("lls") || cmd.equals("ldir")){
            String path=".";
            if(cmds.length==2) {
                path=cmds[1];
            }
            try{
                final java.io.File file=new java.io.File(path);
                if(!file.exists()){
                    out.println(path+": No such file or directory");
                } else
                    if(file.isDirectory()){
                        final String[] list=file.list();
                        for(int ii=0; ii<list.length; ii++){
                            out.println(list[ii]);
                        }
                    } else {
                        out.println(path);
                    }
            } catch(final Exception e){
                err.println(e);
            }
        } else if(cmd.equals("get") ||
                cmd.equals("get-resume") || cmd.equals("get-append") ||
                cmd.equals("put") ||
                cmd.equals("put-resume") || cmd.equals("put-append")) {
            if(cmds.length!=2 && cmds.length!=3) {
                err.println("Invalid arguments");
                return;
            }
            final String p1=cmds[1];
            String p2=".";
            if(cmds.length==3) {
                p2=cmds[2];
            }
            try{
                if(cmd.startsWith("get")){
                    int mode=ChannelSftp.OVERWRITE;
                    if(cmd.equals("get-resume")){ mode=ChannelSftp.RESUME; }
                    else if(cmd.equals("get-append")){ mode=ChannelSftp.APPEND; }
                    channel.get(p1, p2, null, mode);
                }
                else{
                    int mode=ChannelSftp.OVERWRITE;
                    if(cmd.equals("put-resume")){ mode=ChannelSftp.RESUME; }
                    else if(cmd.equals("put-append")){ mode=ChannelSftp.APPEND; }
                    channel.put(p1, p2, null, mode);
                }
            } catch(final SftpException e){
                err.println(e);
            }
        } else if(cmd.equals("ln") || cmd.equals("symlink") || cmd.equals("rename")){
            if(cmds.length!=3) {
                err.println("Invalid arguments");
                return;
            }
            final String p1=cmds[1];
            final String p2=cmds[2];
            try{
                if(cmd.equals("rename")) {
                    channel.rename(p1, p2);
                } else {
                    channel.symlink(p1, p2);
                }
            } catch(final SftpException e){
                err.println(e);
            }
        } else if(cmd.equals("stat") || cmd.equals("lstat")){
            if(cmds.length!=2) {
                err.println("Invalid arguments");
                return;
            }
            final String p1=cmds[1];
            SftpATTRS attrs=null;
            try{
                if(cmd.equals("stat")) {
                    attrs=channel.stat(p1);
                } else {
                    attrs=channel.lstat(p1);
                }
            }
            catch(final SftpException e){
                err.println(e);
            }
            if(attrs!=null){
                out.println(attrs);
            }
        } else if(cmd.equals("readlink")){
            if(cmds.length!=2) {
                err.println("Invalid arguments");
                return;
            }
            final String p1= cmds[1];
            String filename=null;
            try{
                filename=channel.readlink(p1);
                out.println(filename);
            } catch(final SftpException e){
                err.println(e);
            }
        } else if(cmd.equals("realpath")){
            if(cmds.length!=2) {
                err.println("Invalid arguments");
                return;
            }
            final String p1=cmds[1];
            String filename=null;
            try{
                filename=channel.realpath(p1);
                out.println(filename);
            } catch(final SftpException e){
                err.println(e);
            }
        } else if(cmd.equals("version")){
            out.println("SFTP protocol version "+channel.version());
        } else if(cmd.equals("help") || cmd.equals("?")){
            out.println(help);
        } else {
            err.println("unimplemented command: "+cmd);
        }
    }

    private static String help =
        "      Available commands:\n"+
        "      * means unimplemented command.\n"+
        "cd path                       Change remote directory to 'path'\n"+
        "lcd path                      Change local directory to 'path'\n"+
        "chgrp grp path                Change group of file 'path' to 'grp'\n"+
        "chmod mode path               Change permissions of file 'path' to 'mode'\n"+
        "chown own path                Change owner of file 'path' to 'own'\n"+
        "help                          Display this help text\n"+
        "get remote-path [local-path]  Download file\n"+
        "get-resume remote-path [local-path]  Resume to download file.\n"+
        "get-append remote-path [local-path]  Append remote file to local file\n"+
        "*lls [ls-options [path]]      Display local directory listing\n"+
        "ln oldpath newpath            Symlink remote file\n"+
        "*lmkdir path                  Create local directory\n"+
        "lpwd                          Print local working directory\n"+
        "ls [path]                     Display remote directory listing\n"+
        "*lumask umask                 Set local umask to 'umask'\n"+
        "mkdir path                    Create remote directory\n"+
        "put local-path [remote-path]  Upload file\n"+
        "put-resume local-path [remote-path]  Resume to upload file\n"+
        "put-append local-path [remote-path]  Append local file to remote file.\n"+
        "pwd                           Display remote working directory\n"+
        "stat path                     Display info about path\n"+
        "exit                          Quit sftp\n"+
        "quit                          Quit sftp\n"+
        "rename oldpath newpath        Rename remote file\n"+
        "rmdir path                    Remove remote directory\n"+
        "rm path                       Delete remote file\n"+
        "symlink oldpath newpath       Symlink remote file\n"+
        "readlink path                 Check the target of a symbolic link\n"+
        "realpath path                 Canonicalize the path\n"+
        "rekey                         Key re-exchanging\n"+
        "compression level             Packet compression will be enabled\n"+
        "version                       Show SFTP version\n"+
        "?                             Synonym for help";

    public void setVariable(final String name, final String value)
    throws ProtocolException {
        throw new UnsupportedOperationException();
    }

    public void unsetVariable(final String name) throws ProtocolException {
        throw new UnsupportedOperationException();
    }

    public static void main(final String[] arg) throws ProtocolException, IOException {

        final HashMap<String, String> context = new HashMap<String, String>();
        context.put(SshProtocolConfigFactory.PRIVATE_KEY_URL,
                System.getProperty("user.home") + File.separator + ".ssh" + File.separator + "id_rsa");
        context.put(SshProtocolConfigFactory.KNOWN_HOSTS,
                System.getProperty("user.home") + File.separator + ".ssh" + File.separator + "known_hosts");
        context.put(SshProtocolConfigFactory.HOST, arg[0]);
        context.put(SshProtocolConfigFactory.USER, arg[1]);

        final SshProtocolConfig config =
            SshProtocolConfigFactory.createSshProtocolConfig(null, context);
        config.setErrorStream(System.err);
        config.setOutputStream(System.out);

        final SshProtocolSession sshSession = new SshProtocol().createSession(config);
        final SftpShell shell = new SftpShell(sshSession);
        while(shell.channel.isConnected()){
            System.out.print("sftp> ");
            final BufferedReader reader =
                new BufferedReader(new InputStreamReader(System.in));
            final String line = reader.readLine();
            if (line == null) {
                break;
            }
            shell.execute(line);
        }
        FileUtils.close(sshSession);
    }

}
