/**
 * Copyright 2010-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.protocol.ssh.impl.jsch;

import java.util.Map;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.protocol.api.IProtocol;
import org.ow2.util.protocol.api.ProtocolConfig;
import org.ow2.util.protocol.api.ProtocolException;
import org.ow2.util.protocol.ssh.api.SshProtocolConfig;

import com.jcraft.jsch.JSch;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.ProxyHTTP;
import com.jcraft.jsch.ProxySOCKS5;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.UserInfo;

/**
 *
 * @author Loris Bouzonnet
 */
public class SshProtocol implements IProtocol {

    public static final int DEFAULT_SSH_PORT = 22;

    private static Log logger = LogFactory.getLog(SshProtocol.class);

    public SshProtocolSession createSession(final ProtocolConfig config)
    throws ProtocolException {

        final JSch jsch = new JSch();

        try {
            if (config.getPort() == null) {
                config.setPort(DEFAULT_SSH_PORT);
            }

            final Session session =
                jsch.getSession(config.getUser(), config.getHost(), config.getPort());

            String privateKey = null;
            String passphrase = null;

            if (config instanceof SshProtocolConfig) {

                final SshProtocolConfig sshProtocolConfig = (SshProtocolConfig) config;

                privateKey = sshProtocolConfig.getPrivateKeyPath();
                passphrase = sshProtocolConfig.getPassphrase();

                if (privateKey != null) {
                    jsch.addIdentity(
                            privateKey,
                            passphrase);
                }

                if (sshProtocolConfig.getKnownHostsPath() != null) {
                    jsch.setKnownHosts(sshProtocolConfig.getKnownHostsPath());
                }

                if (sshProtocolConfig.getCompressionLevel() != null) {
                    session.setConfig("compression.s2c", "zlib@openssh.com,zlib,none");
                    session.setConfig("compression.c2s", "zlib@openssh.com,zlib,none");
                    session.setConfig("compression_level",
                            sshProtocolConfig.getCompressionLevel().toString());
                }

                final String proxyHttp = sshProtocolConfig.getProxyHttp();
                final String proxySocks4 = sshProtocolConfig.getProxySocks4();
                final String proxySocks5 = sshProtocolConfig.getProxySocks5();
                if (proxyHttp != null && proxySocks4 != null ||
                        proxyHttp != null && proxySocks5 != null ||
                        proxySocks4 != null && proxySocks5 != null) {
                    logger.warn("Cannot simultaneously use many proxies: HTTP will be preferred.");
                }

                if (proxyHttp != null) {
                    session.setProxy(new ProxyHTTP(proxyHttp));
                } else if (proxySocks4 != null) {
                    session.setProxy(new ProxySOCKS5(proxySocks4));
                } else if (proxySocks5 != null) {
                    session.setProxy(new ProxySOCKS5(proxySocks5));
                }

                final String x11Forwarding = sshProtocolConfig.getX11Forwarding();
                if (x11Forwarding != null) {
                    final String[] hostPort = x11Forwarding.split(":");
                    session.setX11Host(hostPort[0]);
                    if (hostPort.length == 2) {
                        session.setX11Port(Integer.parseInt(hostPort[1]) + 6000);
                    }
                }

                if (sshProtocolConfig.getConnectionTimeout() != null) {
                    session.setTimeout(sshProtocolConfig.getConnectionTimeout());
                }

            }

            // username and password will be given via UserInfo interface.
            final UserInfo ui;
            if (config.getPassword() == null && privateKey == null) {
                ui= new InteractiveUserInfo();
            } else {
                ui = new CachedUserInfo(config.getPassword(), passphrase);
            }
            session.setUserInfo(ui);
            session.connect();

            return new SshProtocolSession(session, config);

        } catch (final JSchException e) {
            throw new ProtocolException(e);
        }
    }

    public ProtocolConfig createConfig(final String url, final Map<String, String> context)
    throws ProtocolException {
        return SshProtocolConfigFactory.createSshProtocolConfig(url, context);
    }

}
